/**
 * @copyright nhcarrigan
 * @license Naomi's Public License
 * @author Naomi Carrigan
 */
interface ActivityData {
  id:          number;
  user_id:     number;
  op_type:     string;
  act_user_id: number;
  act_user:    ActUser;
  repo_id:     number;
  repo:        Repo;
  comment_id:  number;
  comment?:    Comment;
  ref_name:    string;
  is_private:  boolean;
  content:     string;
  created:     string;
}

interface ActUser {
  id:                  number;
  login:               string;
  login_name:          string;
  source_id:           number;
  full_name:           string;
  email:               string;
  avatar_url:          string;
  html_url:            string;
  language:            string;
  is_admin:            boolean;
  last_login:          string;
  created:             string;
  restricted:          boolean;
  active:              boolean;
  prohibit_login:      boolean;
  location:            string;
  pronouns:            string;
  website:             string;
  description:         string;
  visibility:          string;
  followers_count:     number;
  following_count:     number;
  starred_repos_count: number;
  username:            string;
}

interface Repo {
  id:                                number;
  owner:                             Owner;
  name:                              string;
  full_name:                         string;
  description:                       string;
  empty:                             boolean;
  private:                           boolean;
  fork:                              boolean;
  template:                          boolean;
  parent:                            unknown;
  mirror:                            boolean;
  size:                              number;
  language:                          string;
  languages_url:                     string;
  html_url:                          string;
  url:                               string;
  link:                              string;
  ssh_url:                           string;
  clone_url:                         string;
  original_url:                      string;
  website:                           string;
  stars_count:                       number;
  forks_count:                       number;
  watchers_count:                    number;
  open_issues_count:                 number;
  open_pr_counter:                   number;
  release_counter:                   number;
  default_branch:                    string;
  archived:                          boolean;
  created_at:                        string;
  updated_at:                        string;
  archived_at:                       string;
  permissions:                       Permissions;
  has_issues:                        boolean;
  external_tracker:                  ExternalTracker;
  has_wiki:                          boolean;
  wiki_branch:                       string;
  globally_editable_wiki:            boolean;
  has_pull_requests:                 boolean;
  has_projects:                      boolean;
  has_releases:                      boolean;
  has_packages:                      boolean;
  has_actions:                       boolean;
  ignore_whitespace_conflicts:       boolean;
  allow_merge_commits:               boolean;
  allow_rebase:                      boolean;
  allow_rebase_explicit:             boolean;
  allow_squash_merge:                boolean;
  allow_fast_forward_only_merge:     boolean;
  allow_rebase_update:               boolean;
  default_delete_branch_after_merge: boolean;
  default_merge_style:               string;
  default_allow_maintainer_edit:     boolean;
  avatar_url:                        string;
  internal:                          boolean;
  mirror_interval:                   string;
  object_format_name:                string;
  mirror_updated:                    string;
  repo_transfer:                     unknown;
  topics:                            unknown;
}

interface Owner {
  id:                  number;
  login:               string;
  login_name:          string;
  source_id:           number;
  full_name:           string;
  email:               string;
  avatar_url:          string;
  html_url:            string;
  language:            string;
  is_admin:            boolean;
  last_login:          string;
  created:             string;
  restricted:          boolean;
  active:              boolean;
  prohibit_login:      boolean;
  location:            string;
  pronouns:            string;
  website:             string;
  description:         string;
  visibility:          string;
  followers_count:     number;
  following_count:     number;
  starred_repos_count: number;
  username:            string;
}

interface Permissions {
  admin: boolean;
  push:  boolean;
  pull:  boolean;
}

interface ExternalTracker {
  external_tracker_url:            string;
  external_tracker_format:         string;
  external_tracker_style:          string;
  external_tracker_regexp_pattern: string;
}

interface Comment {
  id:                 number;
  html_url:           string;
  pull_request_url:   string;
  issue_url:          string;
  user:               unknown;
  original_author:    string;
  original_author_id: number;
  body:               string;
  assets:             Array<unknown>;
  created_at:         string;
  updated_at:         string;
}

interface Issue {
  id:                 number;
  url:                string;
  html_url:           string;
  number:             number;
  user:               Owner;
  original_author:    string;
  original_author_id: number;
  title:              string;
  body:               string;
  ref:                string;
  assets:             Array<unknown>;
  labels:             Array<Label>;
  milestone:          unknown;
  assignee:           unknown;
  assignees:          unknown;
  state:              string;
  is_locked:          boolean;
  comments:           number;
  created_at:         string;
  updated_at:         string;
  closed_at:          unknown;
  due_date:           unknown;
  pull_request?:      PullRequest;
  repository:         Repo;
  pin_order:          number;
}

interface Label {
  id:          number;
  name:        string;
  exclusive:   boolean;
  is_archived: boolean;
  color:       string;
  description: string;
  url:         string;
}

interface PullRequest {
  merged:    boolean;
  merged_at: unknown;
  draft:     boolean;
  html_url:  string;
}

class Codeberg {
  private activityCache:       Array<ActivityData>;
  private readonly issueCache: Array<Issue>;
  private lastUpdate:          Date | null = null;
  public constructor() {
    this.activityCache = [];
    this.issueCache = [];
  }

  public async getActivitiesWithCache(): Promise<Array<ActivityData>> {
    // Stale after 6 hours
    if (this.lastUpdate && Date.now()
    - this.lastUpdate.getTime() < 6 * 60 * 60 * 1000) {
      return this.activityCache;
    }
    return await this.getActivities().then(() => {
      return this.activityCache;
    });
  }

  /* istanbul ignore next -- @preserve */
  public async getIssuesWithCache(): Promise<Array<Issue>> {
    // Stale after 6 hours
    if (this.lastUpdate && Date.now()
    - this.lastUpdate.getTime() < 6 * 60 * 60 * 1000) {
      return this.issueCache;
    }
    return await this.getIssues().then(() => {
      return this.issueCache;
    });
  }

  /* istanbul ignore next -- @preserve */
  private async getIssues(): Promise<void> {
    const repositoryRequest
    = await fetch("https://codeberg.org/api/v1/orgs/nhcarrigan/repos");
    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
    const repositoryData = await repositoryRequest.json() as Array<Repo>;
    for (const repo of repositoryData) {
      const issueRequest
      // eslint-disable-next-line no-await-in-loop
      = await fetch(`https://codeberg.org/api/v1/repos/nhcarrigan/${repo.name}/issues`);
      // eslint-disable-next-line @typescript-eslint/consistent-type-assertions, no-await-in-loop
      const issueData = await issueRequest.json() as Array<Issue>;
      for (const issue of issueData) {
        const isHelpWanted = issue.labels.some((label) => {
          return label.name === "help wanted";
        });
        const isFirsstTimer = issue.labels.some((label) => {
          return label.name === "good first issue";
        });
        if (isHelpWanted || isFirsstTimer) {
          this.issueCache.push(issue);
        }
      }
    }
    this.lastUpdate = new Date();
  }

  private async getActivities(): Promise<void> {
    const activityRequest = await
    // eslint-disable-next-line stylistic/max-len
    fetch("https://codeberg.org/api/v1/users/naomi-lgbt/activities/feeds?only-performed-by=true");
    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
    const activityData = await activityRequest.json() as Array<ActivityData>;
    this.activityCache = activityData;
    this.lastUpdate = new Date();
  }
}

const instantiated = new Codeberg();

/**
 * Fetches and caches the Codeberg activity.
 * @returns An array of activity objects.
 */
const getCodebergActivty = async(): Promise<Array<ActivityData>> => {
  return await instantiated.getActivitiesWithCache();
};

/**
 * Fetches and caches the Codeberg issues.
 * @returns An array of issue objects, filtered for help wanted and good first issues.
 */
const getCodebergIssues = async(): Promise<Array<Issue>> => {
  /* istanbul ignore next -- @preserve */
  return await instantiated.getIssuesWithCache();
};

export { getCodebergActivty, getCodebergIssues };
